@@ -1,8 +1,10 @@
-import copy
 import math
+from pathlib import Path
 
 import diff_match_patch as dmp_module
 import yaml
+from git.objects import base
+
 
 class Blupdate:
     """
@@ -23,7 +25,7 @@ class Blupdate:
     def compute_cctv(self):
         """
         Computes a cctv from self.srcbl to self.dstbl. This will take some effort but should be reasonable with DMP.
-        
+
         The cctv should the a list of triples, each triple being:
           - char start in srcbl (int)
           - char end in srcbl (int)
@@ -58,25 +60,29 @@ class Blupdate:
         # src_chunk is the chunk from source text, represented by mode 0 and -1.
         src_chunk_first_idx = 0
         src_chunk_last_idx = 0
-        
+
         # dst_chunk is the chunk from destination text, represented by mode 0 and 1.
         dst_chunk_first_idx = 0
         dst_chunk_last_idx = 0
 
         for mode, chunk in diffs:
-            if mode == 0: # says the chunk is common, update the src_chunk and dst_chunk first and last indices.
+            if (
+                mode == 0
+            ):  # says the chunk is common, update the src_chunk and dst_chunk first and last indices.
                 src_chunk_first_idx = src_chunk_last_idx
                 dst_chunk_first_idx = dst_chunk_last_idx
                 src_chunk_last_idx += len(chunk)
                 dst_chunk_last_idx += len(chunk)
-            elif mode == -1: # says the chunk is from source text, update src_chunk first and last indices.
+            elif (
+                mode == -1
+            ):  # says the chunk is from source text, update src_chunk first and last indices.
                 src_chunk_first_idx = src_chunk_last_idx
                 src_chunk_last_idx += len(chunk)
-            else: # mode == 1, says the chunks is from destination text, update dst_chunk first and last indices.
+            else:  # mode == 1, says the chunks is from destination text, update dst_chunk first and last indices.
                 dst_chunk_first_idx = dst_chunk_last_idx
                 dst_chunk_last_idx += len(chunk)
 
-            # get char coordinatte (cc) of the common chunk based on source and destination text.    
+            # get char coordinatte (cc) of the common chunk based on source and destination text.
             if mode == 0:
                 src_cc = (src_chunk_first_idx, src_chunk_last_idx)
                 dst_cc = (dst_chunk_first_idx, dst_chunk_last_idx)
@@ -102,16 +108,24 @@ class Blupdate:
           get_cctv_for_coord(9) == (-1, true)
         """
         prev_cct = 0
+        result = None
         for cct in self.cctv:
-            if srcblcoord > cct[0] and srcblcoord < cct[1]: # at inner of the range
-                return (cct[2], True)
-            elif srcblcoord == cct[0] or srcblcoord == cct[1] - 1: # at side of the range
-                return (cct[2], False)
-            elif srcblcoord < cct[0]: # falls between ccts (two range)
-                return (math.ceil((prev_cct + cct[2]) / 2), False)
-            
+            if srcblcoord > cct[0] and srcblcoord < cct[1]:  # at inner of the range
+                result = (cct[2], True)
+            elif (
+                srcblcoord == cct[0] or srcblcoord == cct[1] - 1
+            ):  # at side of the range
+                result = (cct[2], False)
+            elif srcblcoord < cct[0]:  # falls between ccts (two range)
+                result = (math.ceil((prev_cct + cct[2]) / 2), False)
+
+            if result:
+                return result
+
             prev_cct = cct[2]
 
+        if not result:
+            return (-1, False)
 
     def get_context(self, srcblcoord):
         """
@@ -126,11 +140,11 @@ class Blupdate:
 
         # check for left context size less than context_len
         if srcblcoord >= self.context_len:
-            left_context = self.srcbl[srcblcoord-self.context_len:srcblcoord]
+            left_context = self.srcbl[srcblcoord - self.context_len : srcblcoord]
         else:
             left_context = self.srcbl[:srcblcoord]
 
-        right_context = self.srcbl[srcblcoord:srcblcoord+self.context_len]
+        right_context = self.srcbl[srcblcoord : srcblcoord + self.context_len]
         return left_context, right_context
 
     def dmp_find(self, context, dstcoordestimate):
@@ -151,6 +165,8 @@ class Blupdate:
 
         By convention, the function returns -1 when it is unable to compute the new coordinate.
         """
+        if cct == -1:
+            return cct
         context = self.get_context(srcblcoord)
         dstcoordestimate = srcblcoord + cct
         return self.dmp_find(context, dstcoordestimate)
@@ -168,20 +184,97 @@ class Blupdate:
         else:
             return self.get_updated_with_dmp(srcblcoord, cctvforcoord[0])
 
-    def update_layer(self, layer_fn):
-        """
-        Update individual layer, in format {0: [0, 17], 1: [20, 40], ...}
-        """
-        anns = yaml.safe_load(layer_fn.open())
-        for i, ann in anns.items():
-            start_cc = ann[0]
-            end_cc = ann[1]
-            anns[i] = [self.get_updated_coord(start_cc), self.get_updated_coord(end_cc)]
-        yaml.dump(anns, layer_fn.open('w'), default_flow_style=False)
-    
-    def update_annotations(self, opfpath):
+
+def update_span(ann, updater: Blupdate):
+    start = updater.get_updated_coord(ann["span"]["start"])
+    end = updater.get_updated_coord(ann["span"]["end"])
+    if start == -1 and end == -1:
+        ann["span"]["fail_update"] = "both"
+    elif start == -1:
+        ann["span"]["fail_update"] = "start"
+        ann["span"]["end"] == end
+    elif end == -1:
+        ann["span"]["fail_update"] = "end"
+        ann["span"]["start"] == start
+    else:
+        ann["span"]["start"] = start
+        ann["span"]["end"] = end
+
+
+def update_ann_layer(layer, updater: Blupdate):
+    """
+    Update individual layer
+    """
+    for _, ann in layer["annotations"].items():
+        update_span(ann, updater)
+
+
+class PechaBaseUpdate:
+    def __init__(self, src_opf_path, dst_opf_path, context_len=10):
+        self.src_opf_path = Path(src_opf_path)
+        self.dst_opf_path = Path(dst_opf_path)
+        self.context_len = context_len
+
+    @property
+    def index_path(self):
+        return self.dst_opf_path / "index.yml"
+
+    @property
+    def layer_path(self):
+        return self.dst_opf_path / "layers"
+
+    def dump(self, data, output_fn):
+        with output_fn.open("w") as fn:
+            yaml.dump(
+                data, fn, default_flow_style=False, sort_keys=False, allow_unicode=True
+            )
+
+    def load(self, fn):
+        return yaml.safe_load(fn.open(encoding="utf-8"))
+
+    @staticmethod
+    def get_base(opf_path, vol_id):
+        return (opf_path / "base" / f"{vol_id}.txt").read_text(encoding="utf-8")
+
+    def update_layers(self, vol_id, updater):
         """
         Update all the layer annotations
         """
-        for layer_fn in (opfpath/'layers').iterdir():
-            self.update_layer(layer_fn)
\ No newline at end of file
+        for layer_fn in (self.layer_path / vol_id).iterdir():
+            layer = self.load(layer_fn)
+            update_ann_layer(layer, updater)
+            self.dump(layer, layer_fn)
+
+    def update_vol(self, vol_id):
+        src_base = self.get_base(self.src_opf_path, vol_id)
+        dst_base = self.get_base(self.dst_opf_path, vol_id)
+        self.update_layers(
+            vol_id, Blupdate(src_base, dst_base, context_len=self.context_len)
+        )
+
+    def update_text_span(self, span):
+        for span_vol in span:
+            vol_id = span_vol["vol"].split("/")[-1]
+            src_base = self.get_base(self.src_opf_path, vol_id)
+            dst_base = self.get_base(self.dst_opf_path, vol_id)
+            updater = Blupdate(src_base, dst_base, context_len=self.context_len)
+            update_span(span_vol, updater)
+
+    def update_index_layer(self):
+        layer = self.load(self.index_path)
+        for ann in layer["annotations"]:
+            # update text span
+            self.update_text_span(ann["span"])
+
+            # update sub-text span
+            for sub_text in ann["parts"]:
+                self.update_text_span(sub_text["span"])
+
+        self.dump(layer, self.index_path)
+
+    def update(self):
+        for vol_fn in Path(self.dst_opf_path / "base").iterdir():
+            print(f"\t- Updating {vol_fn.stem} ...")
+            self.update_vol(vol_fn.stem)
+        print("[INFO] Updating index ...")
+        self.update_index_layer()
